
const debug = require('debug')('datahub-privilege')
const PRIVILEGE_TYPE = {
    DB_ACCESS: 0,
    SCHEMA_ACCESS: 1,
    TABLE_READ: 2,
    TABLE_INSERT: 3,
    TABLE_UPDATE: 4,
    TABLE_DELETE: 5,
    SQL_MAPPING_EXEC: 6,
    DB_EXECUTE_SQL: 7,
}

class PrivilegeHelper {
    constructor(options) {
        if (typeof options.getPrivileges === 'function') {
            this.getPrivileges = options.getPrivileges
        }
        if (typeof options.getFreeDatabase === 'function') {
            this.getFreeDatabase = options.getFreeDatabase
        }
        this.enablePrivilege = options.enablePrivilege
    }

    canAccessDatabase(database) {
        return this.checkPrivilege(PRIVILEGE_TYPE.DB_ACCESS, database);
    }

    canAccessSchema(database, schema) {
        return this.checkPrivilege(PRIVILEGE_TYPE.SCHEMA_ACCESS, database, schema);
    }

    canAccessTable(database, schema, table) {
        return this.canAccessTableRead(database, schema, table)
    }

    canAccessTableRead(database, schema, table) {
        return this.checkPrivilege(PRIVILEGE_TYPE.TABLE_READ, database, schema, table);
    }

    canAccessTableUpdate(database, schema, table) {
        return this.checkPrivilege(PRIVILEGE_TYPE.TABLE_UPDATE, database, schema, table);
    }

    canAccessTableDelete(database, schema, table) {
        return this.checkPrivilege(PRIVILEGE_TYPE.TABLE_DELETE, database, schema, table);
    }

    canAccessTableInsert(database, schema, table) {
        return this.checkPrivilege(PRIVILEGE_TYPE.TABLE_INSERT, database, schema, table);
    }

    canExecuteSQLMapping(database, schema, sqlKey) {
        return this.checkPrivilege(PRIVILEGE_TYPE.SQL_MAPPING_EXEC, database, schema, sqlKey);
    }

    canExecuteSQL(database) {
        return this.checkPrivilege(PRIVILEGE_TYPE.DB_EXECUTE_SQL, database)
    }

    *init(ctx) {
        this.ctx = ctx
        this.privilege = yield this.getPrivileges(ctx)
        this.freeDatabase = yield this.getFreeDatabase(ctx)
    }

    *getPrivileges(ctx) {
        return {
            databases: {
                ['$some_db$']: {
                    tables: {
                        ['$some_schema$']: {
                            ['$some_table']: 1, // level
                        }
                    },
                    sql_mapping: {
                        ['$some_schema$']: {
                            ['$some_sql_key$']: 1, // level
                        }
                    },
                    sql: 1, // 能否在该实例执行任意 SQL
                }
            },
            // hooks: {},
            user: {
                name: this.state.user || 'default',
                admin: false,
            }
        }
    }

    /**
     * 获取不鉴权的数据库
     *
     * @param {*} ctx koa ctx(this)
     * @returns {Map} 
     * @memberof PrivilegeHelper
     */
    *getFreeDatabase(ctx) {
        return ctx.knexHelper.freeDatabase
    }

    needAuthApi(database) {
        if (this.freeDatabase.has(database)) {
            debug('%s in freedatabase, no needAuthApi', database)
            return false
        }

        return true;
    }

    needAuthUser() {
        if (!this.privilege || !this.privilege.user) {
            debug('warning, privilege empty', this.privilege)
            return false
        }

        // 管理员不需要认证
        if (this.privilege.user.admin) {
            debug('is admin, no needAuthUser')
            return false
        }

        return true
    }

    needAuth(database) {
        if (!this.enablePrivilege) {
            return false
        }

        // 如果是不需要鉴权的用户
        if (!this.needAuthUser()) {
            return false
        }

        // 如果是不需要鉴权的数据库
        if (!this.needAuthApi(database)) {
            return false
        }

        return true
    }

    getTablePrivilege(database, schema, table) {
        if (table) {
            try {
                return this.privilege.databases[database].table[schema][table]
            } catch (e) {
                return null
            }
        }

        if (schema) {
            try {
                debug('getTablePrivilege database', this.privilege.databases[database])
                return this.privilege.databases[database].table[schema]
            } catch (e) {
                return null
            }
        }

        if (database) {
            try {
                return this.privilege.databases[database]
            } catch (e) {
                return null
            }
        }

        return this.privilege
    }

    getSqlMappingPrivilege(database, schema, sqlKey) {
        if (sqlKey) {
            try {
                return this.privilege.databases[database].sql_mapping[schema][sqlKey]
            } catch (e) {
                return null
            }
        }

        if (schema) {
            try {
                return this.privilege.databases[database].sql_mapping[schema]
            } catch (e) {
                return null
            }
        }

        if (database) {
            try {
                return this.privilege.databases[database]
            } catch (e) {
                return null
            }
        }

        return this.privilege
    }
    
    getSqlExecutorPrivilege(database) {
        if (database) {
            try {
                return this.privilege.databases[database].sql
            } catch (e) {
                return 0
            }
        }
        return 0
    }

    checkPrivilege(type, database, schema, tableOrSqlKey) {
        const ret = (() => {
            if (!this.needAuth(database)) {
                return true;
            }

            switch (type) {
                case PRIVILEGE_TYPE.DB_ACCESS:
                    return this.getTablePrivilege(database) != null || this.getSqlMappingPrivilege(database) != null;
                case PRIVILEGE_TYPE.SCHEMA_ACCESS:
                    return this.getTablePrivilege(database, schema) != null || this.getSqlMappingPrivilege(database, schema) != null;
                case PRIVILEGE_TYPE.TABLE_READ:
                    return this.getTablePrivilege(database, schema, tableOrSqlKey) > 0;
                case PRIVILEGE_TYPE.TABLE_INSERT:
                case PRIVILEGE_TYPE.TABLE_UPDATE:
                case PRIVILEGE_TYPE.TABLE_DELETE:
                    return this.getTablePrivilege(database, schema, tableOrSqlKey) > 1;
                case PRIVILEGE_TYPE.SQL_MAPPING_EXEC:
                    return this.getSqlMappingPrivilege(database, schema, tableOrSqlKey) > 0;
                case PRIVILEGE_TYPE.DB_EXECUTE_SQL:
                    return this.getSqlExecutorPrivilege(database) > 0;
                default:
                    return false;
            }
        })()

        debug('checkPrivilege', ret, type, database, schema, tableOrSqlKey)
        return ret
    }
}


function PrivilegeMiddleware(options) {
    return function* (next) {
        this.state.enablePrivilege = options.enablePrivilege
        this.privilegeHelper = new PrivilegeHelper(options)
        if (options.enablePrivilege) {
            yield this.privilegeHelper.init(this)
        }
        yield next
    }
}

module.exports = {
    PrivilegeMiddleware,
    PRIVILEGE_TYPE,
}